/*
  Copyright (C) 2010 DJ Delorie <dj@redhat.com>

  This file is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 3, or (at your option) any later
  version.
  
  This file is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for more details.
  
  You should have received a copy of the GNU General Public License
  along with this file; see the file COPYING3.  If not see
  <http://www.gnu.org/licenses/>.
*/

#include <string.h>
#include "r8c-1b.h"
#include "stick.h"

/* disconnect the RX serial from the USB serial */
#define sgate_off() SGATE = 1

/* connect the RX serial to the USB serial */
#define sgate_on() SGATE = 0

static int flash_mode = 0;

static unsigned char p3port = 0;

static void
p3set (unsigned char mask)
{
  p3port |= mask;
  p3.b = p3port;
}

static void
p3clr (unsigned char mask)
{
  p3port &= ~mask;
  p3.b = p3port;
}

/*****************************************************************************/

#define RXD1 p4.b5
#define TXD1 p3.b7

#define tty_off()	ucon.u1sel = 0, pd3 &= ~USB_RX_MASK
#define tty_rx()	ucon.u1sel = 1
#define tty_txrx()	ucon.u1sel = 3, pd3 |= USB_RX_MASK

void
tty_init ()
{
  /* 8MHz / (9600 * 16) - 1 */
  u1brg = 51;
  u1mr.b = 5;	/* UART, 8N1 */
  u1c0.b = 0;	/* LSB first, CMOS, buffered, f1 */
  u1c1.b = 5;	/* enable tx,rx */
  tty_off();
}

void
tty_putc (char c)
{
  while (!u1c1.ti)
    ;
  u1tb = c;
}

unsigned int
tty_getc ()
{
  while (!u1c1.ri)
    ;
  return u1rb;
}

void
tty_flush ()
{
  unsigned int rv;
  while (u1c1.ri)
    rv = u1rb;
}

void
tty_wait ()
{
  while (!u1c1.ti)
    ;
  while (!u1c0.txept)
    ;
}

/*****************************************************************************/

/* Timer X handler */

timer_init ()
{
  txmr.b = 0;
  tcss.txck = 1; /* f8 = 1 MHz */
  prex = 100;	 /* 10 KHz count */
  tx = 255;
  txmr.txs = 1;

  /*txic = 5;*/
}

void __attribute__((interrupt))
timer_x_handler (void)
{
  /* Called ever 25 msec, or 40 times per sec */
  static int count = 0;
  count ++;
  switch (count & 0x1f) {
  case 0:
    if (flash_mode)
      LED_OFF ();
    else
      LED_ON ();
    break;
  case 1:
    if (flash_mode)
      LED_ON ();
    else
      LED_OFF ();
    break;
  }
}

/* It's 10 counts per actual ms.  This is passed decimilliseconds.  Up
   to 25 mS */
static void
wait_dms (int ms)
{
  byte i;

  while (ms > 128)
    {
      wait_dms (128);
      ms -= 128;
    }

  i = tx;
  while ((unsigned char)((char)i - (char)tx) < ms)
    ;
}

/*****************************************************************************/

break_detected ()
{
  int r;

  RESET_ON ();

  /* wait for end of break */
  while (!RXD1)
    ;

  /* connect us to the serial bus */
  sgate_off ();
  tty_txrx ();
  tty_flush ();

  r = tty_getc ();
  if (r & 1)
    {
      flash_mode = 1;
      p3clr (MD1_MASK); /* flashing mode */
    }
  else
    {
      flash_mode = 0;
      p3set (MD1_MASK); /* running mode */
    }

  tty_putc (r ^ 0x20);

  tty_wait ();

  sgate_on ();
  tty_off ();

  RESET_OFF ();
}


/* The logic is as follows:

   UART is configured for 9600 baud

   we look for a LOW on the Rx port of 2ms or more.  That's 20 counts
   on Timer X

   We read one byte at 9600 baud.  The LSB is SET for flashing mode,
   or CLEAR for running mode.  We send the same byte back, with bit 5
   inverted, then release the serial bus again.

*/

main()
{
  byte i;
  int cache[50];
  int cp = 0;
  unsigned int j, p;
  int b, r0, r1;

  RESET_ON ();

  pd1 = RESET_MASK | SGATE_MASK;
  pd3 = LED_MASK | MD1_MASK;

  p3port = 0;

  p3clr (LED_MASK);
  p3set (MD1_MASK);
  LED_ON ();
  sgate_on ();

  timer_init ();

  tty_init ();


  RESET_OFF ();

  while (1)
    {
      if (TXD1 == flash_mode)
	LED_ON ();
      else
	LED_OFF ();
	

      if (RXD1)
	{
	  /* This is the normal idle state, as well as some bits in
	     regular transmissions.  */
	  i = tx;
	}
      else
	{
	  int l = (unsigned char)(i - tx);
	  /* Active state.  If this persists long enough, it's a break.  */
	  if (l > 20)
	    {
	      /* Break */
	      break_detected ();
	    }
	  else
	    {
	      /* Not yet */
	    }

	}
    }
}
